future_month_list = ['F', 'G', 'H', 'J', 'K', 'M',
                     'N', 'Q', 'U', 'V', 'X', 'Z']

short_month = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
               'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']


def future_date_to_code(dte):
    """Obtain a future code from a date.

    For example december 2010 will result in Z10.
    """
    return '%s%s' % (future_month_list[dte.month-1], str(dte.year)[2:])


future_month_dict = dict((future_month_list[i],
                          (i+1, short_month[i])) for i in range(0, 12))



from datetime import date


class BaseHoliday(object):

    def __init__(self, description=''):
        self.description = description

    def extended(self):
        return None

    def allholidays(self, year):
        return []

    def __repr__(self):
        d = self.extended()
        if d:
            return '%s: %s' % (self.__class__.__name__, d)
        else:
            return self.__class__.__name__


class PartialDate(BaseHoliday):

    def __init__(self, month, day):
        self.month = month
        self.day = day

    def allholidays(self, year):
        return (date(year, self.month, self.day),)



import time
from datetime import datetime, date

try:
    from dateutil.parser import parse as date_from_string
except ImportError: # noqa

    def date_from_string(dte):
        raise NotImplementedError


def todate(val):
    '''Convert val to a datetime.date instance by trying several
    conversion algorithm.
    If it fails it raise a ValueError exception.
    '''
    if not val:
        raise ValueError("Value not provided")
    if isinstance(val, datetime):
        return val.date()
    elif isinstance(val, date):
        return val
    else:
        try:
            ival = int(val)
            sval = str(ival)
            if len(sval) == 8:
                return yyyymmdd2date(val)
            elif len(sval) == 5:
                return juldate2date(val)
            else:
                raise ValueError
        except Exception:
            # Try to convert using the parsing algorithm
            try:
                return date_from_string(val).date()
            except Exception:
                raise ValueError("Could not convert %s to date" % val)


def date2timestamp(dte):
    return time.mktime(dte.timetuple())


def jstimestamp(dte):
    '''Convert a date to a javascript timestamp.

    A Javascript timestamp is the number of milliseconds since
    January 1, 1970 00:00:00 UTC.'''
    return 1000*date2timestamp(dte)


def timestamp2date(tstamp):
    "Converts a unix timestamp to a Python datetime object"
    dt = datetime.fromtimestamp(tstamp)
    if not dt.hour+dt.minute+dt.second+dt.microsecond:
        return dt.date()
    else:
        return dt


def yyyymmdd2date(dte):
    try:
        y = dte // 10000
        md = dte % 10000
        m = md // 100
        d = md % 100
        return date(y, m, d)
    except Exception:
        raise ValueError('Could not convert %s to date' % dte)


def date2yyyymmdd(dte):
    return dte.day + 100*(dte.month + 100*dte.year)


def juldate2date(val):
    '''Convert from a Julian date/datetime to python date or datetime'''
    ival = int(val)
    dec = val - ival
    try:
        val4 = 4*ival
        yd = val4 % 1461
        st = 1899
        if yd >= 4:
            st = 1900
        yd1 = yd - 241
        y = val4 // 1461 + st
        if yd1 >= 0:
            q = yd1 // 4 * 5 + 308
            qq = q // 153
            qr = q % 153
        else:
            q = yd // 4 * 5 + 1833
            qq = q // 153
            qr = q % 153
        m = qq % 12 + 1
        d = qr // 5 + 1
    except Exception:
        raise ValueError('Could not convert %s to date' % val)
    if dec:
        dec24 = 24*dec
        hours = int(dec24)
        minutes = int(60*(dec24 - hours))
        tot_seconds = 60*(60*(dec24 - hours) - minutes)
        seconds = int(tot_seconds)
        microseconds = int(1000000*(tot_seconds-seconds))
        return datetime(y, m, d, hours, minutes, seconds, microseconds)
    else:
        return date(y, m, d)


def date2juldate(val):
    '''Convert from a python date/datetime to a Julian date & time'''
    f = 12*val.year + val.month - 22803
    fq = f // 12
    fr = f % 12
    dt = (fr*153 + 302)//5 + val.day + fq*1461//4
    if isinstance(val, datetime):
        return dt + (val.hour + (val.minute + (
            val.second + 0.000001*val.microsecond)/60.)/60.)/24.
    else:
        return dt



'''Day Counter for Counting time between 2 dates.
Implemented::

    * Actual 360
    * Actual 365
    * 30 / 360
    * Actual Actual
'''
from copy import copy
from datetime import date


__all__ = ['getdc', 'ActActYears', 'alldc']


def getdc(name):
    dc = _day_counters.get(name)
    if dc:
        return dc()
    else:
        return None


def alldc():
    global _day_counters
    return copy(_day_counters)


def ActActYears(dt):
    y = dt.year
    r = y % 4
    a = 0.0
    if r > 0:
        a = 1.0
    dd = (dt - date(y, 1, 1)).days
    return y + dd/(365.0 + a)


class DayCounterMeta(type):

    def __new__(cls, name, bases, attrs):
        new_class = super(DayCounterMeta, cls).__new__(cls, name, bases, attrs)
        if new_class.name:
            _day_counters[new_class.name] = new_class
        return new_class


_day_counters = {}


class DayCounter(object):
    name = None
    __metaclass__ = DayCounterMeta

    def count(self, start, end):
        return (end-start).days

    def dcf(self, start, end):
        return self.count(start, end)/360.0


class act360(DayCounter):
    name = 'ACT/360'


class act365(DayCounter):
    name = 'ACT/365'

    def dcf(self, start, end):
        return self.count(start, end)/365.0


class thirty360(DayCounter):
    name = '30/360'

    def count(self, start, end):
        d1 = min(start.day, 30)
        d2 = min(end.day, 30)
        return 360*(end.year - start.year) + 30*(end.month -
                                                 start.month) + d2 - d1


class actact(DayCounter):
    name = 'ACT/ACT'

    def dcf(self, start, end):
        return ActActYears(end) - ActActYears(start)



from .centres import centres, TradingCentre

__all__ = ['centres', 'TradingCentre', 'prevbizday', 'nextbizday']


def prevbizday(dte=None, nd=1, tcs=None):
    tcs = centres(tcs)
    return tcs.prevbizday(dte, nd)


def nextbizday(dte=None, nd=1, tcs=None):
    tcs = centres(tcs)
    return tcs.nextbizday(dte, nd)



"""Python currencies"""

__version__ = '1.0.0'


from .core.currency import (
    currency, currencydb, ccypair, currency_pair,
    dump_currency_table
)
from .core.country import (
    country, countryccy, set_new_country,
    countries, set_country_map, country_map,
    CountryError, eurozone, print_eurozone
)
from .core.daycounter import getdc, ActActYears, alldc
from .dates.converters import (
    todate, date2timestamp, timestamp2date, yyyymmdd2date,
    date2yyyymmdd, juldate2date, date2juldate, date_from_string,
    jstimestamp
)
from .dates.futures import future_date_to_code, future_month_dict
from .dates.period import period, Period


__all__ = [
    'currency', 'currencydb', 'ccypair', 'currency_pair',
    'dump_currency_table',
    #
    'getdc', 'ActActYears', 'alldc',
    #
    'country', 'countryccy', 'set_new_country',
    'countries', 'set_country_map', 'country_map',
    'CountryError', 'eurozone', 'print_eurozone',
    'future_date_to_code', 'future_month_dict',
    'period', 'Period',
    'todate', 'date2timestamp', 'timestamp2date',
    'yyyymmdd2date', 'date2yyyymmdd', 'juldate2date',
    'date2juldate', 'date_from_string', 'jstimestamp'
]


# Shortcuts
def cross(code):
    return currency(code).as_cross()


def crossover(code):
    return currency(code).as_cross('/')


def all():
    return currencydb().keys()


def g7():
    return ['EUR', 'GBP', 'USD', 'CAD']


def g10():
    return g7() + ['CHF', 'SEK', 'JPY']


def g10m():
    """modified g10 = G10 + AUD, NZD, NOK
    """
    return g10() + ['AUD', 'NZD', 'NOK']



#
# Requires pytz 2008i or higher
#
from .currency import currencydb

# Eurozone countries (officially the euro area)
# see http://en.wikipedia.org/wiki/Eurozone
# using ISO 3166-1 alpha-2 country codes
# see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
#
eurozone = tuple(('AT BE CY DE EE ES FI FR GR IE IT LU LV LT MT '
                  'NL PT SI SK').split(' '))


def print_eurozone():
    for c in sorted(map(country, eurozone)):
        print(c)


_countries = None
_country_ccys = None
_country_maps = {}


class CountryError(Exception):
    pass


def country(code):
    cdb = countries()
    code = country_map(code)
    return cdb.get(code, '')


def countryccy(code):
    cdb = countryccys()
    code = str(code).upper()
    return cdb.get(code, None)


def countries():
    '''
    get country dictionar from pytz and add some extra.
    '''
    global _countries
    if not _countries:
        v = {}
        _countries = v
        try:
            from pytz import country_names
            for k, n in country_names.items():
                v[k.upper()] = n
        except Exception:
            pass
    return _countries


def countryccys():
    '''
    Create a dictionary with keys given by countries ISO codes and values
    given by their currencies
    '''
    global _country_ccys
    if not _country_ccys:
        v = {}
        _country_ccys = v
        ccys = currencydb()
        for c in eurozone:
            v[c] = 'EUR'
        for c in ccys.values():
            if c.default_country:
                v[c.default_country] = c.code
    return _country_ccys


def set_country_map(cfrom, cto, name=None, replace=True):
    '''
    Set a mapping between a country code to another code
    '''
    global _country_maps
    cdb = countries()
    cfrom = str(cfrom).upper()
    c = cdb.get(cfrom)
    if c:
        if name:
            c = name
        cto = str(cto).upper()
        if cto in cdb:
            raise CountryError('Country %s already in database' % cto)
        cdb[cto] = c
        _country_maps[cfrom] = cto
        ccys = currencydb()
        cccys = countryccys()
        ccy = cccys[cfrom]
        cccys[cto] = ccy

        # If set, remove cfrom from database
        if replace:
            ccy = ccys.get(ccy)
            ccy.default_country = cto
            cdb.pop(cfrom)
            cccys.pop(cfrom)
    else:
        raise CountryError('Country %s not in database' % c)


def set_new_country(code, ccy, name):
    '''
    Add new country code to database
    '''
    code = str(code).upper()
    cdb = countries()
    if code in cdb:
        raise CountryError('Country %s already in database' % code)
    ccys = currencydb()
    ccy = str(ccy).upper()
    if ccy not in ccys:
        raise CountryError('Currency %s not in database' % ccy)
    cdb[code] = str(name)
    cccys = countryccys()
    cccys[code] = ccy


def country_map(code):
    '''
    Country mapping
    '''
    code = str(code).upper()
    global _country_maps
    return _country_maps.get(code, code)


# Add eurozone to list of Countries
set_new_country('EU', 'EUR', 'Eurozone')



import datetime
# from dateutil import rrule

from .holiday import BaseHoliday, PartialDate


# weekdays = (rrule.MO,rrule.TU,rrule.WE,rrule.TH,rrule.FR)
# weekend  = (rrule.SA,rrule.SU)
isoweekend = (6, 7)
oneday = datetime.timedelta(days=1)

_tcs = {}


def centres(codes=None):
    tcs = TradingCentres()
    if codes:
        lcs = codes.upper().replace(' ', '').split(',')
        for code in lcs:
            tc = _tcs.get(code)
            if tc:
                tcs.add(tc)
    return tcs


def get_declared_holidays(bases, attrs):
    _attrs = attrs.copy()
    holidays = [(field_name, attrs.pop(field_name)) for field_name, obj in
                _attrs.items() if isinstance(obj, BaseHoliday)]

    for base in bases[::-1]:
        if hasattr(base, 'holidays'):
            holidays = base.holidays.items() + holidays

    return dict(holidays)


class TradingCentreMeta(type):

    def __new__(cls, name, bases, attrs):
        global _tcs
        abstract = attrs.pop('abstract', False)
        if abstract:
            return super(TradingCentreMeta, cls).__new__(cls, name,
                                                         bases, attrs)
        else:
            attrs['holidays'] = get_declared_holidays(bases, attrs)
            new_class = super(TradingCentreMeta, cls).__new__(cls, name,
                                                              bases, attrs)
            if not getattr(new_class, 'abstract', False):
                cl = new_class()
                _tcs[cl.code] = cl
            return new_class


TradingCentreBase = TradingCentreMeta('TradingCentreBase',
                                      (object, ),
                                      {'abstract': True})


class TradingCentre(TradingCentreBase):
    abstract = True
    onedaydelta = datetime.timedelta(days=1)

    def __new__(cls):
        obj = super(TradingCentre, cls).__new__(cls)
        obj._start = None
        obj._end = None
        obj._cache = {}
        return obj

    def __get_code(self):
        return self.__class__.__name__
    code = property(__get_code)

    def isbizday(self, dte):
        if dte.isoweekday() in isoweekend:
            return False
        else:
            return self._isholiday(dte) is False

    def _isholiday(self, dte):
        year = dte.year
        if self._start:
            if year >= self._start and year <= self._end:
                return self._cache.get(dte, False)
            elif year < self._start:
                start = year
                end = self._start - 1
            else:
                start = self._end + 1
                end = year
        else:
            start = year
            end = year
        for year in range(start, end+1):
            self.build_dates(year)
        return self._cache.get(dte, False)

    def build_dates(self, year):
        for holiday in self.holidays.values():
            days = holiday.allholidays(year)
            for day in days:
                self._cache[day] = True


class TradingCentres(object):
    onedaydelta = datetime.timedelta(days=1)

    def __init__(self):
        self._centres = {}

    def __len__(self):
        return len(self._centres)

    def add(self, tc):
        self._centres[tc.code] = tc

    def _isbizday(self, dte):
        for c in self._centres.values():
            if c._isholiday(dte):
                return False
        return True

    def isbizday(self, dte):
        if dte.isoweekday() in isoweekend:
            return False
        return self._isbizday(dte)

    def nextbizday(self, dte, nd=1):
        n = 0
        isbz = self.isbizday
        while not isbz(dte):
            dte += oneday
        while n < nd:
            dte += oneday
            if isbz(dte):
                n += 1
        return dte

    def prevbizday(self, dte, nd=1):
        n = 0
        if nd < 0:
            return self.nextbizday(dte, -nd)
        else:
            while not self.isbizday(dte):
                dte -= self.onedaydelta
            n = 0
            while n < nd:
                dte -= self.onedaydelta
                if self.isbizday(dte):
                    n += 1
        return dte


#
# TRADING CENTRES
class TGT(TradingCentre):
    '''Target'''
    new_years_day = PartialDate(1, 1)
    labor_day = PartialDate(5, 1)
    christmas_day = PartialDate(12, 25)
    december_26 = PartialDate(12, 26)


class LON(TradingCentre):
    '''London'''
    christmas_day = PartialDate(12, 25)



import os
from setuptools import setup, find_packages

import ccy


def read(name):
    filename = os.path.join(os.path.dirname(__file__), name)
    with open(filename, encoding='utf8') as fp:
        return fp.read()


def requirements(name):
    install_requires = []
    dependency_links = []

    for line in read(name).split('\n'):
        if line.startswith('-e '):
            link = line[3:].strip()
            if link == '.':
                continue
            dependency_links.append(link)
            line = link.split('=')[1]
        line = line.strip()
        if line:
            install_requires.append(line)

    return install_requires, dependency_links


install_requires = requirements('dev/requirements.txt')[0]
tests_require = requirements('dev/requirements-dev.txt')[0]


meta = dict(
    name='ccy',
    version=ccy.__version__,
    description=ccy.__doc__,
    author='Luca Sbardella',
    author_email="luca@quantmind.com",
    maintainer_email="luca@quantmind.com",
    url="https://github.com/lsbardel/ccy",
    license="BSD",
    long_description=read('README.rst'),
    packages=find_packages(include=['ccy', 'ccy.*']),
    install_requires=install_requires,
    tests_require=tests_require,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Office/Business :: Financial',
        'Topic :: Utilities'
    ]
)


if __name__ == '__main__':
    setup(**meta)



def period(pstr=''):
    '''Create a period object from a period string'''
    return Period.make(pstr)


def find_first_of(st, possible):
    possible = tuple(possible)
    lowi = -1
    for p in possible:
        i = st.find(p)
        if i != -1 and (i < lowi or lowi == -1):
            lowi = i
    return lowi


def safediv(x, d):
    return x // d if x >= 0 else -(-x // d)


def safemod(x, d):
    return x % d if x >= 0 else -(-x % d)


class Period:

    def __init__(self, months=0, days=0):
        self._months = months
        self._days = days

    @classmethod
    def make(cls, pstr=''):
        if isinstance(pstr, cls):
            return pstr
        else:
            return cls().add_tenure(pstr)

    def isempty(self):
        return self._months == 0 and self._days == 0

    def add_days(self, days):
        self._days += days

    def add_weeks(self, weeks):
        self._days += int(7*weeks)

    def add_months(self, months):
        self._months += months

    def add_years(self, years):
        self._months += int(12*years)

    @property
    def years(self):
        return safediv(self._months, 12)

    @property
    def months(self):
        return safemod(self._months, 12)

    @property
    def weeks(self):
        return safediv(self._days, 7)

    @property
    def days(self):
        return safemod(self._days, 7)

    @property
    def totaldays(self):
        return 30*self._months + self._days

    def __repr__(self):
        '''The period string'''
        return self.components()

    def __str__(self):
        return self.__repr__()

    def components(self):
        '''The period string'''
        p = ''
        neg = self.totaldays < 0
        y = self.years
        m = self.months
        w = self.weeks
        d = self.days
        if y:
            p = '%sY' % abs(y)
        if m:
            p = '%s%sM' % (p, abs(m))
        if w:
            p = '%s%sW' % (p, abs(w))
        if d:
            p = '%s%sD' % (p, abs(d))
        return '-'+p if neg else p

    def simple(self):
        '''A string representation with only one period delimiter.'''
        if self._days:
            return '%sD' % self.totaldays
        elif self.months:
            return '%sM' % self._months
        elif self.years:
            return '%sY' % self.years
        else:
            return ''

    def add_tenure(self, pstr):
        if isinstance(pstr, self.__class__):
            self._months += pstr._months
            self._days += pstr._days
            return self
        st = str(pstr).upper()
        done = False
        sign = 1
        while not done:
            if not st:
                done = True
            else:
                ip = find_first_of(st, 'DWMY')
                if ip == -1:
                    raise ValueError("Unknown period %s" % pstr)
                p = st[ip]
                v = int(st[:ip])
                sign = sign if v > 0 else -sign
                v = sign*abs(v)
                if p == 'D':
                    self.add_days(v)
                elif p == 'W':
                    self.add_weeks(v)
                elif p == 'M':
                    self.add_months(v)
                elif p == 'Y':
                    self.add_years(v)
                st = st[ip+1:]
        return self

    def __add__(self, other):
        other = self.make(other)
        return self.__class__(self._months+other._months,
                              self._days+other._days)

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        other = self.make(other)
        return self.__class__(self._months-other._months,
                              self._days-other._days)

    def __rsub__(self, other):
        return self.make(other) - self

    def __gt__(self, other):
        return self.totaldays > self.make(other).totaldays

    def __lt__(self, other):
        return self.totaldays < self.make(other).totaldays

    def __ge__(self, other):
        return self.totaldays >= self.make(other).totaldays

    def __le__(self, other):
        return self.totaldays <= self.make(other).totaldays

    def __eq__(self, other):
        return self.totaldays == self.make(other).totaldays



import sys

from .data import make_ccys


usd_order = 5


def to_string(v):
    if isinstance(v, bytes):
        return v.decode('utf-8')
    else:
        return '%s' % v


def overusdfun(v1):
    return v1


def overusdfuni(v1):
    return 1./v1


class ccy(object):
    '''
    Currency object
    '''
    def __init__(self, code, isonumber, twoletterscode, order, name,
                 roundoff=4,
                 default_country=None,
                 fixeddc=None,
                 floatdc=None,
                 fixedfreq=None,
                 floatfreq=None,
                 future=None,
                 symbol=r'\00a4',
                 html=''):
        self.code = to_string(code)
        self.id = self.code
        self.isonumber = isonumber
        self.twoletterscode = to_string(twoletterscode)
        self.order = int(order)
        self.name = to_string(name)
        self.rounding = roundoff
        self.default_country = default_country
        self.symbol_raw = symbol
        self.symbol = symbol.encode('utf-8').decode('unicode_escape')
        self.html = html or self.symbol
        self.fixeddc = fixeddc
        self.floatdc = floatdc
        self.future = ''
        if future:
            self.future = str(future)

    def __getstate__(self):
        return {'code': self.code}

    def __setstate__(self, dict):
        c = currency(dict['code'])
        self.__dict__.update(c.__dict__)

    def __eq__(self, other):
        if isinstance(other, ccy):
            return other.code == self.code
        return False

    def description(self):
        if self.order > usd_order:
            v = 'USD / %s' % self.code
        else:
            v = '%s / USD' % self.code
        if self.order != usd_order:
            return '%s Spot Exchange Rate' % v
        else:
            return 'Dollar'

    def info(self):
        return {'code': self.code,
                'isonumber': self.isonumber,
                'twoletterscode': self.twoletterscode,
                'symbol': self.symbol,
                'order': self.order,
                'name': self.name,
                'rounding': self.rounding,
                'default_country': self.default_country,
                'unicode symbol': self.symbol_raw}

    def printinfo(self, stream=None):
        info = self.info()
        stream = stream or sys.stdout
        for k, v in info.items():
            stream.write(to_string('%s: %s\n' % (k, v)))

    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__, self.code)

    def __str__(self):
        return self.code

    def swap(self, c2):
        '''
        put the order of currencies as market standard
        '''
        inv = False
        c1 = self
        if c1.order > c2.order:
            ct = c1
            c1 = c2
            c2 = ct
            inv = True
        return inv, c1, c2

    def overusdfunc(self):
        if self.order > usd_order:
            return overusdfuni
        else:
            return overusdfun

    def usdoverfunc(self):
        if self.order > usd_order:
            return overusdfun
        else:
            return overusdfuni

    def as_cross(self, delimiter=''):
        '''
        Return a cross rate representation with respect USD.
        @param delimiter: could be '' or '/' normally
        '''
        if self.order > usd_order:
            return 'USD%s%s' % (delimiter, self.code)
        else:
            return '%s%sUSD' % (self.code, delimiter)

    def spot(self, c2, v1, v2):
        if self.order > c2.order:
            vt = v1
            v1 = v2
            v2 = vt
        return v1/v2


class ccy_pair(object):
    '''
    Currency pair such as EURUSD, USDCHF

    XXXYYY - XXX is the foreign currency, while YYY is the base currency

    XXXYYY means 1 unit of of XXX cost XXXYYY units of YYY
    '''
    def __init__(self, c1, c2):
        self.ccy1 = c1
        self.ccy2 = c2
        self.code = '%s%s' % (c1, c2)
        self.id = self.code

    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__, self.code)

    def __str__(self):
        return self.code

    def mkt(self):
        if self.ccy1.order > self.ccy2.order:
            return ccy_pair(self.ccy2, self.ccy1)
        else:
            return self

    def over(self, name='usd'):
        '''Returns a new currency pair with the *over* currency as
        second part of the pair (Foreign currency).'''
        name = name.upper()
        if self.ccy1.code == name.upper():
            return ccy_pair(self.ccy2, self.ccy1)
        else:
            return self


class ccydb(dict):

    def insert(self, *args, **kwargs):
        c = ccy(*args, **kwargs)
        self[c.code] = c


def currencydb():
    global _ccys
    if not _ccys:
        _ccys = ccydb()
        make_ccys(_ccys)
    return _ccys


def ccypairsdb():
    global _ccypairs
    if not _ccypairs:
        _ccypairs = make_ccypairs()
    return _ccypairs


def currency(code):
    c = currencydb()
    return c.get(str(code).upper())


def ccypair(code):
    c = ccypairsdb()
    return c.get(str(code).upper())


def currency_pair(code):
    '''Construct a :class:`ccy_pair` from a six letter string.'''
    c = str(code)
    c1 = currency(c[:3])
    c2 = currency(c[3:])
    return ccy_pair(c1, c2)


def make_ccypairs():
    ccys = currencydb()
    db = {}

    for ccy1 in ccys.values():
        od = ccy1.order
        for ccy2 in ccys.values():
            if ccy2.order <= od:
                continue
            p = ccy_pair(ccy1, ccy2)
            db[p.code] = p
    return db


def dump_currency_table():
    headers = ['code',
               'name',
               ('isonumber', 'iso'),
               ('html', 'symbol'),
               ('default_country', 'country'),
               'order',
               'rounding']
    all = []
    data = []
    all.append(data)
    for h in headers:
        if isinstance(h, tuple):
            h = h[1]
        data.append(h)
    for c in sorted(currencydb().values(), key=lambda x: x.order):
        data = []
        all.append(data)
        for h in headers:
            if isinstance(h, tuple):
                h = h[0]
            data.append(getattr(c, h))
    return all


_ccys = None
_ccypairs = None




def make_ccys(db):
    '''
    Create the currency dictionary
    '''
    dfr = 4
    dollar = r'\u0024'
    peso = r'\u20b1'
    kr = r'kr'
    insert = db.insert

    # G10 & SCANDI
    insert('EUR', '978', 'EU', 1,
           'Euro', dfr, 'EU', '30/360', 'ACT/360',
           future='FE', symbol=r'\u20ac', html='&#x20ac;')
    insert('GBP', '826', 'BP', 2,
           'British Pound', dfr, 'GB', 'ACT/365', 'ACT/365',
           symbol=r'\u00a3', html='&#xa3;')
    insert('AUD', '036', 'AD', 3,
           'Australian Dollar', dfr, 'AU', 'ACT/365', 'ACT/365',
           symbol=dollar, html='&#x24;')
    insert('NZD', '554', 'ND', 4,
           'New-Zealand Dollar', dfr, 'NZ', 'ACT/365', 'ACT/365',
           symbol=dollar, html='&#x24;')
    insert('USD', '840', 'UD', 5,
           'US Dollar', 0, 'US', '30/360', 'ACT/360',
           future='ED', symbol=dollar, html='&#x24;')
    insert('CAD', '124', 'CD', 6,
           'Canadian Dollar', dfr, 'CA', 'ACT/365', 'ACT/365',
           symbol=dollar, html='&#x24;')
    insert('CHF', '756', 'SF', 7,
           'Swiss Franc', dfr, 'CH', '30/360', 'ACT/360',
           symbol=r'Fr', html='&#x20a3;')
    insert('NOK', '578', 'NK', 8,
           'Norwegian Krona', dfr, 'NO', '30/360', 'ACT/360',
           symbol=kr, html=kr)
    insert('SEK', '752', 'SK', 9,
           'Swedish Krona', dfr, 'SE', '30/360', 'ACT/360',
           symbol=kr, html=kr)
    insert('DKK', '208', 'DK', 10,
           'Danish Krona', dfr, 'DK', '30/360', 'ACT/360',
           symbol=kr, html=kr)
    insert('JPY', '392', 'JY', 10000,
           'Japanese Yen', 2,  'JP', 'ACT/365', 'ACT/360',
           symbol=r'\u00a5', html='&#xa5;')

    # ASIA
    insert('CNY', '156', 'CY', 680,
           'Chinese Renminbi', dfr, 'CN', 'ACT/365', 'ACT/365',
           symbol=r'\u00a5', html='&#xa5;')
    insert('KRW', '410', 'KW', 110000,
           'South Korean won', 2, 'KR', 'ACT/365', 'ACT/365',
           symbol=r'\u20a9', html='&#x20a9;')
    insert('SGD', '702', 'SD', 15,
           'Singapore Dollar', dfr, 'SG', 'ACT/365', 'ACT/365',
           symbol=dollar, html='&#x24;')
    insert('IDR', '360', 'IH', 970000,
           'Indonesian Rupiah', 0, 'ID', 'ACT/360', 'ACT/360',
           symbol=r'Rp', html='Rp')
    insert('THB', '764', 'TB', 3300,
           'Thai Baht', 2, 'TH', 'ACT/365', 'ACT/365',
           symbol=r'\u0e3f', html='&#xe3f;')
    insert('TWD', '901', 'TD', 18,
           'Taiwan Dollar', dfr, 'TW', 'ACT/365', 'ACT/365',
           symbol=dollar, html='&#x24;')
    insert('HKD', '344', 'HD', 19,
           'Hong Kong Dollar', dfr, 'HK', 'ACT/365', 'ACT/365',
           symbol=r'\u5713', html='HK&#x24;')
    insert('PHP', '608', 'PP', 4770,
           'Philippines Peso', dfr, 'PH', 'ACT/360', 'ACT/360',
           symbol=peso, html='&#x20b1;')
    insert('INR', '356', 'IR', 4500,
           'Indian Rupee', dfr, 'IN', 'ACT/365', 'ACT/365',
           symbol=r'\u20a8', html='&#x20a8;')
    insert('MYR', '458', 'MR', 345,
           'Malaysian Ringgit', dfr, 'MY', 'ACT/365', 'ACT/365')
    insert('VND', '704', 'VD', 1700000,
           'Vietnamese Dong', 0,  'VN', 'ACT/365', 'ACT/365',
           symbol=r'\u20ab', html='&#x20ab;')

    # LATIN AMERICA
    insert('BRL', '986', 'BC', 200,
           'Brazilian Real', dfr, 'BR', 'BUS/252', 'BUS/252',
           symbol=r'R$')
    insert('PEN', '604', 'PS', 220,
           'Peruvian New Sol', dfr, 'PE', 'ACT/360', 'ACT/360',
           symbol=r'S/.')
    insert('ARS', '032', 'AP', 301,
           'Argentine Peso', dfr, 'AR', '30/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    insert('MXN', '484', 'MP', 1330,
           'Mexican Peso', dfr, 'MX', 'ACT/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    insert('CLP', '152', 'CH', 54500,
           'Chilean Peso', 2,  'CL', 'ACT/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    insert('COP', '170', 'CL', 190000,
           'Colombian Peso', 2,  'CO', 'ACT/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    # TODO: Check towletters code and position
    insert('JMD', '388', 'JD', 410,
           'Jamaican Dollar', dfr, 'JM', 'ACT/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    # TODO: Check towletters code and position
    insert('TTD', '780', 'TT', 410,
           'Trinidad and Tobago Dollar', dfr, 'TT', 'ACT/360', 'ACT/360',
           symbol=dollar, html='&#x24;')
    # TODO: Check towletters code and position
    insert('BMD', '060', 'BD', 410,
           'Bermudian Dollar', dfr, 'BM',
           symbol=dollar, html='&#x24;')

    # EASTERN EUROPE
    insert('CZK', '203', 'CK', 28,
           'Czech Koruna', dfr, 'CZ', 'ACT/360', 'ACT/360',
           symbol=r'\u004b\u010d')
    insert('PLN', '985', 'PZ', 29,
           'Polish Zloty', dfr, 'PL', 'ACT/ACT', 'ACT/365',
           symbol=r'\u0050\u0142')
    insert('TRY', '949', 'TY', 30,
           'Turkish Lira', dfr, 'TR', 'ACT/360', 'ACT/360',
           symbol=r'\u0054\u004c')
    insert('HUF', '348', 'HF', 32,
           'Hungarian Forint', dfr, 'HU', 'ACT/365', 'ACT/360',
           symbol=r'Ft', html='Ft')
    insert('RON', '946', 'RN', 34,
           'Romanian Leu', dfr, 'RO', 'ACT/360', 'ACT/360')
    insert('RUB', '643', 'RR', 36,
           'Russian Ruble', dfr, 'RU', 'ACT/ACT', 'ACT/ACT',
           symbol=r'\u0440\u0443\u0431')
    # TODO: Check towletters code and position
    insert('HRK', '191', 'HK', 410,
           'Croatian kuna', dfr, 'HR',
           symbol=r'kn')
    # TODO: Check towletters code and position
    insert('KZT', '398', 'KT', 410,
           'Tenge',  dfr, 'KZ',
           symbol=r'\u20b8', html='&#x20b8;')
    # TODO: Check towletters code and position
    insert('BGN', '975', 'BN', 410,
           'Bulgarian Lev', dfr, 'BG',
           symbol=r'\u043b\u0432.', html='&#1083;&#1074;')

    # MIDDLE EAST & AFRICA
    insert('ILS', '376', 'IS', 410,
           'Israeli Shekel', dfr, 'IL', 'ACT/365', 'ACT/365',
           symbol=r'\u20aa', html='&#x20aa;')
    # TODO: Check towletters code and position
    insert('AED', '784', 'AE', 410,
           'United Arab Emirates Dirham', dfr, 'AE')
    # TODO: Check towletters code and position
    insert('QAR', '634', 'QA', 410,
           'Qatari Riyal', dfr, 'QA',
           symbol=r'\ufdfc', html='&#xfdfc;')
    # TODO: Check towletters code and position
    insert('SAR', '682', 'SR', 410,
           'Saudi Riyal', dfr, 'SA',
           symbol=r'\ufdfc', html='&#xfdfc;')
    insert('EGP', '818', 'EP', 550,
           'Egyptian Pound', dfr, 'EG',
           symbol=r'\u00a3', html='&#xa3;')
    insert('ZAR', '710', 'SA', 750,
           'South African Rand', dfr, 'ZA', 'ACT/365', 'ACT/365',
           symbol=r'R', html='R')

    # BITCOIN
    insert('XBT', '000', 'BT', -1,
           'Bitcoin', 8, 'WW',
           symbol=r'\u0e3f', html='&#xe3f;')
